/******************************************************************************\
 *  _   ___     ____  __               _                                      *
 * | | | \ \___/ /  \/  | ___ __ _ ___| |_                                    *
 * | |_| |\     /| |\/| |/ __/ _` / __| __|                                   *
 * |  _  | \ - / | |  | | (_| (_| \__ \ |_                                    *
 * |_| |_|  \_/  |_|  |_|\___\__,_|___/\__|                                   *
 *                                                                            *
 * This file is part of the HAMcast project.                                  *
 *                                                                            *
 * HAMcast is free software: you can redistribute it and/or modify            *
 * it under the terms of the GNU Lesser General Public License as published   *
 * by the Free Software Foundation, either version 3 of the License, or       *
 * (at your option) any later version.                                        *
 *                                                                            *
 * HAMcast is distributed in the hope that it will be useful,                 *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of             *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                       *
 * See the GNU Lesser General Public License for more details.                *
 *                                                                            *
 * You should have received a copy of the GNU Lesser General Public License   *
 * along with HAMcast. If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                            *
 * Contact: HAMcast support <hamcast-support@informatik.haw-hamburg.de>       *
\******************************************************************************/

#include <map>
#include <set>
#include <list>
#include <memory>
#include <algorithm>

#include <boost/thread.hpp>

#include "hamcast/config.hpp"
#include "hamcast/membership_event.hpp"

#include "hamcast/util/sink.hpp"
#include "hamcast/util/source.hpp"
#include "hamcast/util/future.hpp"
#include "hamcast/util/socket_io.hpp"
#include "hamcast/util/read_buffer.hpp"
#include "hamcast/util/id_generator.hpp"
#include "hamcast/util/buffered_sink.hpp"
#include "hamcast/util/serialization.hpp"
#include "hamcast/util/buffered_source.hpp"

#include "hamcast/ipc/client_channel.hpp"
#include "hamcast/ipc/middleware_configuration.hpp"

#include <iostream>
using std::cerr;
using std::endl;

namespace {

using namespace hamcast;
using namespace hamcast::ipc;
using namespace hamcast::detail;

boost::mutex m_instance_mtx;
client_channel::ptr m_instance;

std::pair<util::source::ptr, util::sink::ptr> new_client_socket_io()
{
    middleware_configuration mc;
    mc.read();
    util::socket_io::ptr sio = util::socket_io::create("127.0.0.1", mc.port());
    std::pair<util::source::ptr, util::sink::ptr> result;
    result.first.reset(sio.get());
    result.second.reset(new util::buffered_sink<>(sio));
    // verify this client; lifetime scope of s
    {
        util::serializer s(result.second);
        s << magic_number
          << major_version
          << minor_version;
    }
    // accepted?
    {
        boost::uint8_t answer;
        util::deserializer d(result.first);
        d >> answer;
        if (answer != 1)
        {
            throw connection_to_middleware_failed(incompatible_middleware_found);
        }
    }
    return result;
}

struct outgoing_mcast_packet : client_channel_msg
{
    char* content;
    socket_id sck_id;
    stream_id stm_id;
    boost::uint32_t content_size;
    sequence_number seq_nr;
    outgoing_mcast_packet() : client_channel_msg(omp_t), content(0) { }
    outgoing_mcast_packet(socket_id sock,
                          stream_id stream,
                          boost::uint32_t ct_size,
                          char* ct)
        : client_channel_msg(omp_t), content(ct)
        , sck_id(sock), stm_id(stream), content_size(ct_size)
    {
    }
    virtual ~outgoing_mcast_packet()
    {
        if (content) delete[] content;
    }
    virtual void* addr() { return this; }
};

typedef outgoing_mcast_packet* omp_ptr;

struct incoming_mcast_packet : client_channel_msg
{
    socket_id sck_id;
    hamcast::multicast_packet mcast_packet;
    incoming_mcast_packet(socket_id sid, const hamcast::multicast_packet& mp)
        : client_channel_msg(imp_t), sck_id(sid), mcast_packet(mp)
    {
    }
    virtual void* addr() { return this; }
};

struct outgoing_synchronous_message : client_channel_msg
{

    function_id fid;
    const util::const_buffer* buf;
    util::future<message::ptr>* fmsg;

    outgoing_synchronous_message(function_id sync_fid,
                                 const util::const_buffer* cbuf,
                                 util::future<message::ptr>* msg_future)
        : client_channel_msg(osm_t), fid(sync_fid), buf(cbuf), fmsg(msg_future)
    {
        HC_REQUIRE(cbuf && msg_future);
    }
    virtual void* addr() { return this; }

};

struct incoming_synchronous_message : client_channel_msg
{
    message::ptr msg;
    inline incoming_synchronous_message(const message::ptr& outgoing_msg)
        : client_channel_msg(ism_t)
        , msg(outgoing_msg)
    {
        HC_REQUIRE(outgoing_msg.get() != 0);
    }
    virtual void* addr() { return this; }
};

struct reg_d_ptr : client_channel_msg
{
    socket_id sck_id;
    multicast_socket::d_ptr ptr;
    inline reg_d_ptr(socket_id sid, const multicast_socket::d_ptr& dp)
        : client_channel_msg(reg_d_ptr_t), sck_id(sid), ptr(dp)
    {
    }
    virtual void* addr() { return this; }
};

struct ack_or_rtm_msg : client_channel_msg
{
    stream_id stm_id;
    socket_id sck_id;
    sequence_number seq_nr;
    inline ack_or_rtm_msg(client_channel_msg_type mt,
                          stream_id v1, socket_id v2, sequence_number v3)
        : client_channel_msg(mt), stm_id(v1), sck_id(v2), seq_nr(v3)
    {
        HC_REQUIRE(mt == rtm_t || mt == ack_t);
    }
    virtual void* addr() { return this; }
};

struct event_msg : client_channel_msg
{
    membership_event event;
    event_msg(const membership_event& mev)
        : client_channel_msg(event_t), event(mev)
    {
    }
    virtual void* addr() { return this; }
};

struct register_callback_msg : client_channel_msg
{
    membership_event_callback fun;
    register_callback_msg(const membership_event_callback& cb_fun)
        : client_channel_msg(reg_cb_t), fun(cb_fun)
    {
    }
    virtual void* addr() { return this; }
};

struct omp_ptr_less
{
    inline bool operator()(const omp_ptr& lhs, const omp_ptr& rhs)
    {
        // nullptr < {everything}
        return (lhs && rhs) ? lhs->seq_nr < rhs->seq_nr : rhs != 0;
    }
};

typedef std::map<sequence_number, omp_ptr> omp_ptr_map;

typedef std::map<socket_id, omp_ptr_map> socket_mcast_buf_map;

enum socket_state_t
{
    uninitialized_state,
    socket_open_state,
    socket_about_to_close_state,
    socket_closed_state
};

struct output_stream
{
    long pending_sends;
    hamcast::ipc::sequence_number last_sent;
    util::id_generator<hamcast::ipc::sequence_number> seq_nrs;
    omp_ptr_map buffer;
    inline output_stream() : pending_sends(0), last_sent(0) { }
};

template<typename T>
struct scope_cleaner { };

template<>
struct scope_cleaner<omp_ptr_map>
{
    omp_ptr_map& mbuf;
    scope_cleaner(omp_ptr_map& buf) : mbuf(buf) { }
    ~scope_cleaner()
    {
        for (omp_ptr_map::iterator i = mbuf.begin(); i != mbuf.end(); ++i)
        {
            delete (i->second);
        }
        mbuf.clear();
    }
};

template<>
struct scope_cleaner<socket_mcast_buf_map>
{
    socket_mcast_buf_map& mb_map;
    scope_cleaner(socket_mcast_buf_map& scoped_map) : mb_map(scoped_map) { }
    ~scope_cleaner()
    {
        for (socket_mcast_buf_map::iterator k = mb_map.begin();
             k != mb_map.end();
             ++k)
        {
            scope_cleaner<omp_ptr_map> tmp(k->second);
        }
        mb_map.clear();
    }
};

// returns: the number of released bytes of all content fields
// with respect to max_buffered_sends (each content_size is counted as
// max(content_size, default_max_buffer_size / max_buffered_sends)
boost::uint32_t drop_smaller_or_equal_seq_nrs(output_stream& ostm,
                                             boost::uint32_t seq_nr)
{
    HC_LOG_TRACE("seq_nr = " << seq_nr);
    omp_ptr_map& mb = ostm.buffer;
    boost::uint32_t result = 0;
    omp_ptr_map::iterator begin = mb.begin();
    // end = first element whose key is > seq_nr
    omp_ptr_map::iterator end = mb.upper_bound(seq_nr);
    HC_LOG_DEBUG("head: " << begin->first << ", upper_bound = " << (end == mb.end() ? 0 : end->first));
    omp_ptr_map tmp;
    long removed_items = 0;
    if (begin != end)
    {
        tmp.insert(begin, end);
        scope_cleaner<omp_ptr_map> cleaner(tmp);
        for (omp_ptr_map::iterator i = begin; i != end; ++i)
        {
//			result += i->second->content_size;
            result += std::max(i->second->content_size, static_cast<boost::uint32_t>(min_buffer_chunk));
            ++removed_items;
        }
        mb.erase(begin, end);
        // cleaner deletes all dropped pointers
    }
    HC_LOG_DEBUG(removed_items << " elements removed");
    begin = mb.begin();
    // end = first element whose key is > seq_nr
    if (begin == mb.end())
    {
        HC_LOG_DEBUG("new head: -empty-");
    }
    else
    {
        HC_LOG_DEBUG("new head: " << begin->first);
    }
    ostm.pending_sends -= removed_items;
    return result;
}

void do_send(util::serializer& s, output_stream& ostm,
             ipc::sequence_number from, ipc::sequence_number to)
{
    HC_LOG_TRACE("from = " << from << ", to = " << to);
    long sent = 0;
    omp_ptr_map::iterator begin = ostm.buffer.find(from);
    if (begin == ostm.buffer.end())
    {
//		not an error; just nothing to send
//		HC_LOG_ERROR(from << " not found");
        return;
    }
    omp_ptr_map::iterator end = ostm.buffer.upper_bound(to);
    for (omp_ptr_map::iterator i = begin; i != end; ++i)
    {
        omp_ptr o = i->second;
        HC_LOG_DEBUG("send seq_nr " << o->seq_nr);
        message::serialize_as_message(s,
                                      async_send,
                                      o->stm_id,
                                      o->sck_id,
                                      o->seq_nr,
                                      o->content_size,
                                      o->content);
        ++sent;
    }
    s.flush();
    ostm.pending_sends += sent;
    ostm.last_sent += sent;
}

struct socket_data
{
    socket_state_t state;
    outgoing_synchronous_message* delayed_close_message;
    hamcast::multicast_socket::d_ptr mcast_obj;
    std::map<hamcast::ipc::stream_id, output_stream> out_streams;
    socket_data() : state(uninitialized_state), delayed_close_message(0) { }
    bool has_pending_sends()
    {
        if (out_streams.empty())
        {
            return false;
        }
        else
        {
            for (std::map<hamcast::ipc::stream_id, output_stream>::iterator i = out_streams.begin();
                 i != out_streams.end();
                 ++i)
            {
                if (!(i->second.buffer.empty())) return true;
            }
            return false;
        }
    }
};

typedef std::map<socket_id,socket_data> socket_data_map;
typedef std::map<request_id, outgoing_synchronous_message*> requests_map;

} // namespace <anonymous>

namespace hamcast { namespace detail {

client_channel_msg::~client_channel_msg() { }

} } // namespace hamcast::detail

namespace hamcast { namespace ipc {

client_channel::ptr client_channel::get()
{
    boost::mutex::scoped_lock guard(m_instance_mtx);
    if (!m_instance)
    {
        m_instance.reset(new client_channel());
        m_instance->run();
    }
    return m_instance;
}

client_channel::client_channel() : channel(new_client_socket_io())
{
}

void client_channel::register_d_ptr(socket_id sid,
                                    const multicast_socket::d_ptr &ptr)
{
    m_queue.push(new reg_d_ptr(sid, ptr));
}

void client_channel::register_callback(const membership_event_callback& cb)
{
    m_queue.push(new register_callback_msg(cb));
}

// wiretap create and delete operations
void wiretap(socket_data_map& smap,
             outgoing_synchronous_message* request,
             const sync_response_view& resp_view)
{
    if (resp_view.exc_id() != eid_none) return;
    switch (request->fid)
    {

     case fid_create_socket:
        {
            HC_REQUIRE(resp_view.content());
            socket_id sid;
            // lifetime scope of d
            {
                util::const_buffer cb(resp_view.content_size(),
                                      resp_view.content());
                util::deserializer d(new util::read_buffer(cb));
                d >> sid;
            }
            HC_REQUIRE(smap.insert(std::make_pair(sid, socket_data())).second);
        }
        break;

     case fid_create_send_stream:
        {
            HC_REQUIRE(resp_view.content() && request->buf);
            // fetch returned stream id
            stream_id stm_id;
            // lifetime scope of d
            {
                util::const_buffer cb(resp_view.content_size(),
                                      resp_view.content());
                util::deserializer d(new util::read_buffer(cb));
                d >> stm_id;
            }
            // fetch socket id from request
            socket_id sck_id;
            {
                util::const_buffer cb2(request->buf->size(),
                                       request->buf->data());
                util::deserializer d2(new util::read_buffer(cb2));
                d2 >> sck_id;
            }
            socket_data_map::iterator i = smap.find(sck_id);
            HC_REQUIRE(i != smap.end());
            HC_REQUIRE(i->second.out_streams.insert(std::make_pair(stm_id, output_stream())).second);
        }
        break;

     case fid_delete_socket:
        {
            HC_REQUIRE(request->buf);
            // fetch socket id from request
            socket_id sck_id;
            {
                util::const_buffer cb2(request->buf->size(),
                                       request->buf->data());
                util::deserializer d2(new util::read_buffer(cb2));
                d2 >> sck_id;
            }
            socket_data_map::iterator i = smap.find(sck_id);
            HC_REQUIRE(i != smap.end());
            i->second.state = socket_closed_state;
        }
        break;

     default: break;

    }
}

void client_channel::send_loop()
{
    HC_LOG_TRACE("");
    util::id_generator<request_id> req_ids;
    socket_data_map m_sockets;
    client_channel_msg* e = 0;
    util::serializer s(m_sink);
    requests_map m_requests;
    std::list<membership_event_callback> m_cbs;
    for (;;)
    {
        e = m_queue.pop();
        switch (e->type)
        {

         case reg_d_ptr_t:
            {
                HC_LOG_SCOPE("reg_d_ptr_t", "");
                reg_d_ptr* rdp = dynamic_cast<reg_d_ptr*>(e);
                m_sockets[rdp->sck_id].mcast_obj = rdp->ptr;
                delete rdp;
            }
            break;

         case omp_t:
            {
                HC_LOG_SCOPE("omp_t", "");
                outgoing_mcast_packet* omp = dynamic_cast<outgoing_mcast_packet*>(e);
                socket_data_map::iterator i = m_sockets.find(omp->sck_id);
                if (i == m_sockets.end())
                {
                    HC_LOG_ERROR("socket " << omp->sck_id << " not found");
                    delete omp;
                }
                else
                {
                    socket_data& sdata = i->second;
                    output_stream& ostm = sdata.out_streams[omp->stm_id];
                    omp->seq_nr = ostm.seq_nrs.next();
                    HC_LOG_DEBUG("seq_nr = " << omp->seq_nr);
                    if (omp->seq_nr == 0 || omp->seq_nr == (ostm.last_sent + 1))
                    {
                        if (ostm.pending_sends < max_pending_sends)
                        {
                            HC_LOG_DEBUG("send seq_nr " << omp->seq_nr);
                            ostm.pending_sends += 1;
                            message::serialize_as_message(s,
                                                          async_send,
                                                          omp->stm_id,
                                                          omp->sck_id,
                                                          omp->seq_nr,
                                                          omp->content_size,
                                                          omp->content);
                            s.flush();
                            ostm.last_sent = omp->seq_nr;
                        }
                    }
                    ostm.buffer.insert(std::make_pair(omp->seq_nr, omp));
                }
            }
            break;

         case rtm_t:
            {
                ack_or_rtm_msg* rmsg = dynamic_cast<ack_or_rtm_msg*>(e);
                HC_LOG_SCOPE("rtm_t", "seq_nr = " << rmsg->seq_nr);
                cerr << "RETRANSMIT" << endl;
                socket_data_map::iterator i = m_sockets.find(rmsg->sck_id);
                if (i != m_sockets.end())
                {
                    socket_data& sdata = i->second;
                    output_stream& ostm = sdata.out_streams[rmsg->stm_id];
                    sequence_number from = rmsg->seq_nr;
                    sequence_number to = ostm.last_sent;
                    do_send(s, ostm, from, to);
                }
                delete rmsg;
            }
            break;

         case ack_t:
            {
                ack_or_rtm_msg* amsg = e->downcast<ack_or_rtm_msg>();
                HC_LOG_SCOPE("ack_t", "socket = " << amsg->sck_id
                                      << ", seq_nr = " << amsg->seq_nr);
                socket_data_map::iterator i = m_sockets.find(amsg->sck_id);
                if (i != m_sockets.end())
                {
                    socket_data& sdata = i->second;
                    output_stream& ostm = sdata.out_streams[amsg->stm_id];
                    boost::uint32_t released = drop_smaller_or_equal_seq_nrs(ostm, amsg->seq_nr);
                    multicast_socket::release_storage(sdata.mcast_obj, released);
                    // make sure there are still pending sends
                    if (ostm.pending_sends < max_pending_sends)
                    {
                        sequence_number from = ostm.last_sent + 1;
                        sequence_number to = from + (max_pending_sends - ostm.pending_sends);
                        do_send(s, ostm, from, to);
                    }
                    if (sdata.state == socket_about_to_close_state)
                    {
                        if (   !sdata.has_pending_sends()
                            && sdata.delayed_close_message)
                        {
                            cerr << "delayed close message send" << endl;
                            m_queue.push(sdata.delayed_close_message);
                            sdata.delayed_close_message = 0;
                        }
                    }
                }
                else
                {
                    HC_LOG_ERROR("ACK on an unknown socket");
                }
                delete amsg;
            }
            break;

         case imp_t:
            {
                incoming_mcast_packet* imp = e->downcast<incoming_mcast_packet>();
                HC_LOG_SCOPE("imp_t", "sck_id = " << imp->sck_id);
                socket_data_map::iterator i = m_sockets.find(imp->sck_id);
                if (i == m_sockets.end())
                {
                    HC_LOG_ERROR("socket " << imp->sck_id << " not found");
                }
                else
                {
                    socket_data& sdata = i->second;
                    multicast_socket::async_recv(sdata.mcast_obj,
                                                 imp->mcast_packet);
                }
                delete imp;
            }
            break;

         case osm_t:
            {
                outgoing_synchronous_message* osm = e->downcast<outgoing_synchronous_message>();
                HC_LOG_SCOPE("osm_t", "fid = " << osm->fid);
                bool do_not_send = false;
                if (osm->fid == fid_delete_socket)
                {
                    socket_id sck_id;
                    util::const_buffer mcb(osm->buf->size(),
                                           osm->buf->data());
                    util::deserializer md(new util::read_buffer(mcb));
                    md >> sck_id;
                    socket_data_map::iterator i = m_sockets.find(sck_id);
                    HC_REQUIRE(i != m_sockets.end());
                    socket_data& sdata = i->second;
                    if (sdata.has_pending_sends())
                    {
                        sdata.delayed_close_message = osm;
                        sdata.state = socket_about_to_close_state;
                        // do not send message now
                        do_not_send = true;
                    }
                }
                if (!do_not_send)
                {
                    request_id rid = req_ids.next();
                    message::serialize_as_message(s,
                                                  sync_request,
                                                  osm->fid,
                                                  rid,
                                                  0,
                                                  static_cast<boost::uint32_t>(osm->buf->size()),
                                                  osm->buf->data());
                    s.flush();
                    m_requests[rid] = osm;
                }
            }
            break;

         case ism_t:
            {
                incoming_synchronous_message* ism = e->downcast<incoming_synchronous_message>();
                sync_response_view srv(*(ism->msg));
                HC_LOG_SCOPE("ism_t", "rid = " << srv.req_id());
                requests_map::iterator i = m_requests.find(srv.req_id());
                if (i == m_requests.end())
                {
                    HC_LOG_ERROR("request " << srv.req_id() << " not found");
                }
                else
                {
                    outgoing_synchronous_message* osm = i->second;
                    m_requests.erase(i);
                    wiretap(m_sockets, osm, srv);
                    osm->fmsg->set(ism->msg);
                    delete osm;
                }
                delete ism;
            }
            break;

         case event_t:
            {
                event_msg* emsg = e->downcast<event_msg>();
                for (std::list<membership_event_callback>::iterator i = m_cbs.begin();
                     i != m_cbs.end();
                     ++i)
                {
                    (*i)(emsg->event);
                }
                delete emsg;
            }
            break;

         case reg_cb_t:
            {
                m_cbs.push_back(e->downcast<register_callback_msg>()->fun);
                delete e;
            }
            break;

         default:
            throw std::logic_error("invalid message type");

        }
    }
}

struct recv_helper
{

    enum state_type
    {
        rd_msg_type,
        rd_field1,
        rd_field2,
        rd_field3,
        rd_content_size,
        rd_content,
        done
    }
    state;

    char* content;

    boost::uint16_t msg_type_field;
    boost::uint16_t field1;
    boost::uint32_t field2;
    boost::uint32_t field3;
    boost::uint32_t content_size;

    util::read_buffer rd_buf;

    recv_helper() : state(rd_msg_type), content(0)
    {
    }

    template<typename T>
    inline void try_read(T& what, bool& finished, state_type next_state)
    {
        if (rd_buf.available() >= sizeof(T))
        {
            rd_buf.read(sizeof(T), &what);
            state = next_state;
        }
        else
        {
            finished = true;
        }
    }

    char* take_content()
    {
        char* result = content;
        content = 0;
        return result;
    }

    void reset()
    {
        state = rd_msg_type;
        if (content)
        {
            delete[] content;
            content = 0;
        }
    }

    state_type operator()()
    {
        bool finished = false;
        while (!finished)
        {
            switch (state)
            {

             case rd_msg_type:
                try_read(msg_type_field, finished, rd_field1);
                break;

             case rd_field1:
                try_read(field1, finished, rd_field2);
                break;

             case rd_field2:
                try_read(field2, finished, rd_field3);
                break;

             case rd_field3:
                try_read(field3, finished, rd_content_size);
                break;

             case rd_content_size:
                try_read(content_size, finished, rd_content);
                break;

             case rd_content:
                if (content_size == 0)
                {
                    state = done;
                }
                else if (rd_buf.available() >= content_size)
                {
                    content = new char[content_size];
                    rd_buf.read(content_size, content);
                    state = done;
                }
                finished = true;
                break;

             default: throw std::logic_error("invalid state");

            }
        }
        return state;
    }

    ~recv_helper()
    {
        if (content) delete[] content;
    }

    // compile to a queue message
    client_channel_msg* compile()
    {
        HC_REQUIRE(state == done);
        client_channel_msg* result = 0;
        message_type mtype = static_cast<message_type>(msg_type_field);
        switch (mtype)
        {

         case async_recv:
            {
                HC_REQUIRE(content_size > 0 && content);
                util::deserializer d(new util::read_buffer(content_size,
                                                           content, true));
                content = 0; // d owns the buffer now
                multicast_packet mp;
                d >> mp;
                result = new incoming_mcast_packet(
                        static_cast<socket_id>(field2), mp);
            }
            break;

         case sync_response:
            {
                // message takes ownership of content
                result = new incoming_synchronous_message(
                                message::create(mtype, field1, field2, field3,
                                                content_size, content));
                content = 0;
            }
            break;

         case retransmit:
         case cumulative_ack:
            result = new ack_or_rtm_msg((mtype == retransmit ? rtm_t : ack_t),
                                        field1, field2, field3);
            break;

         case async_event:
            {
                HC_REQUIRE(content_size > 0 && content);
                util::deserializer d(new util::read_buffer(content_size,
                                                           content, true));
                content = 0; // d owns the buffer now
                uri grp;
                d >> grp;
                if (grp.empty())
                {
                    throw std::logic_error("empty group name");
                }
                // verify event type
                switch (static_cast<membership_event_type>(field1))
                {
                 case join_event:
                 case leave_event:
                 case new_source_event:
                    break;
                 default: throw std::logic_error("invalid event type");
                }
                membership_event_type met
                        = static_cast<membership_event_type>(field1);
                interface_id iid = static_cast<interface_id>(field2);
                result = new event_msg(membership_event(grp, iid, met));
            }
            break;

         default:
            throw std::logic_error("invalid message type");

        }
        reset();
        return result;
    }

};

void client_channel::receive_loop()
{
    util::write_buffer<> wb;
    wb.reserve(default_max_write_buffer_size);
    recv_helper rh;
    client_channel_msg* incomings_head = 0;
    client_channel_msg* incomings_tail = 0;
    try
    {
        for (;;)
        {
            // init (get some fresh data)
            wb.append_some_from(*m_source, wb.remaining());
            rh.rd_buf.reset(wb.size(), wb.data());
            while (rh() == recv_helper::done)
            {
                client_channel_msg* ccm = rh.compile();
                if (ccm->type == imp_t)
                {
                    ccm->next = incomings_head;
                    incomings_head = ccm;
                    if (!incomings_tail) incomings_tail = ccm;
                }
                else
                {
                    m_queue.push(ccm);
                }
            }
            // enqueue all incoming messages at once
            if (incomings_head)
            {
                if (incomings_head == incomings_tail)
                {
                    m_queue.push(incomings_head);
                }
                else
                {
                    m_queue.push(incomings_head, incomings_tail);
                }
                incomings_head = incomings_tail = 0;
            }
            // cleanup (erase read data from buffer)
            wb.erase(0, rh.rd_buf.read_position());
        }
    }
    catch (std::exception& e)
    {
        cerr << "exception: " << e.what() << endl;
        exit(42);
    }
}

void client_channel::on_exit(const std::string& err_str)
{
    throw connection_to_middleware_lost(err_str);
}

void client_channel::send(const message::ptr& what)
{
    throw std::runtime_error("Not implemented ... :P");
}

void client_channel::send_async_data(socket_id sock, boost::uint16_t stream,
                                     boost::uint32_t ct_size, char* ct)
{
    outgoing_mcast_packet* pck = new outgoing_mcast_packet(sock, stream,
                                                           ct_size, ct);
    // make sure constructor is done
    HC_MEMORY_BARRIER();
    m_queue.push(pck);
}

message::ptr client_channel::send_sync_request(function_id fid,
                                               const util::const_buffer& buf)
{
    util::future<message::ptr> fmsg;
    outgoing_synchronous_message* msg = new outgoing_synchronous_message(fid,
                                                                         &buf,
                                                                         &fmsg);
    // make sure constructor is done
    HC_MEMORY_BARRIER();
    m_queue.push(msg);
    return fmsg.get();
}

} } // namespace hamcast::ipc
