/******************************************************************************\
 *  _   ___     ____  __               _                                      *
 * | | | \ \___/ /  \/  | ___ __ _ ___| |_                                    *
 * | |_| |\     /| |\/| |/ __/ _` / __| __|                                   *
 * |  _  | \ - / | |  | | (_| (_| \__ \ |_                                    *
 * |_| |_|  \_/  |_|  |_|\___\__,_|___/\__|                                   *
 *                                                                            *
 * This file is part of the HAMcast project.                                  *
 *                                                                            *
 * HAMcast is free software: you can redistribute it and/or modify            *
 * it under the terms of the GNU Lesser General Public License as published   *
 * by the Free Software Foundation, either version 3 of the License, or       *
 * (at your option) any later version.                                        *
 *                                                                            *
 * HAMcast is distributed in the hope that it will be useful,                 *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of             *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                       *
 * See the GNU Lesser General Public License for more details.                *
 *                                                                            *
 * You should have received a copy of the GNU Lesser General Public License   *
 * along with HAMcast. If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                            *
 * Contact: HAMcast support <hamcast-support@informatik.haw-hamburg.de>       *
\******************************************************************************/

#include <boost/thread/mutex.hpp>
#include <boost/lexical_cast.hpp>
#include "hamcast/util/socket_io.hpp"
#include "hamcast/util/buffered_sink.hpp"
#include "hamcast/util/buffered_source.hpp"

#ifdef WIN_32
#else
#include <arpa/inet.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <fcntl.h>
#endif

#include <iostream>
using std::cout;
using std::cerr;
using std::endl;

namespace {

class native_socket_io : public hamcast::util::socket_io
{

    // stores closed state
    volatile bool m_closed;

    // native socket descriptor
    hamcast::util::native_socket m_sockfd;

    int m_mss;

    static int read_mss(int sockfd)
    {
        int result;
        socklen_t mss_len = sizeof(result);
        getsockopt(sockfd, IPPROTO_TCP, TCP_MAXSEG, &result, &mss_len);
        return result;
    }

    static bool set_tcp_nodelay(hamcast::util::native_socket sockfd)
    {
        int flag = 1;
        // set TCP_NODELAY
        return setsockopt(sockfd, IPPROTO_TCP, TCP_NODELAY,
                          &flag, sizeof(int)) == 0;
    }

 public:

    native_socket_io(const char* host_ip, boost::uint16_t hp) : m_closed(true)
    {
        HC_REQUIRE(host_ip != NULL);
        m_sockfd = socket(AF_INET, SOCK_STREAM, 0);
        if (!set_tcp_nodelay(m_sockfd))
        {
            HC_LOG_WARN("Unable to set TCP_NODELAY");
        }
        sockaddr_in serv_addr;
        memset(reinterpret_cast<void*>(&serv_addr), 0, sizeof(sockaddr_in));
        serv_addr.sin_family = AF_INET;
        serv_addr.sin_addr.s_addr = inet_addr(host_ip);
        serv_addr.sin_port = htons(hp);
        // connect to server
        if (0 != connect(m_sockfd, reinterpret_cast<sockaddr*>(&serv_addr),
                         sizeof(sockaddr_in)))
        {
            std::string err = "Could not connect to ";
            err += host_ip;
            err += " on port ";
            err += boost::lexical_cast<std::string>(hp);
            HC_LOG_FATAL(err);
            throw std::ios_base::failure(err);
        }
        else m_closed = false;
        m_mss = read_mss(m_sockfd);
    }

    native_socket_io(hamcast::util::native_socket ns)
        : m_closed(false), m_sockfd(ns)
    {
        if (!set_tcp_nodelay(m_sockfd))
        {
            HC_LOG_WARN("Unable to set TCP_NODELAY");
        }
        m_mss = read_mss(m_sockfd);
    }

    ~native_socket_io()
    {
        close();
    }

    bool wait_for_data(long sec, long usec)
    {
        HC_REQUIRE(!closed());
        struct timeval tv;
        tv.tv_sec = sec;
        tv.tv_usec = usec;
        fd_set rfds;
        FD_ZERO(&rfds);
        FD_SET(m_sockfd, &rfds);
        int retval = select(m_sockfd + 1, &rfds, 0, 0, &tv);
        HC_REQUIRE(retval != -1);
        return retval != 0;
    }

    virtual int flush_hint() const
    {
        return m_mss;
    }

    void write(size_t buf_size, const void* buf)
    {
        HC_LOG_TRACE("buf_size: " << buf_size);
        ssize_t bsize = static_cast<ssize_t>(buf_size);
        HC_REQUIRE(bsize >= 0 && buf != NULL);
        if (!m_closed)
        {
            if (bsize != send(m_sockfd, buf, buf_size, 0))
            {
                std::string err = "Cannot write to socket";
                HC_LOG_DEBUG(err);
                throw std::ios_base::failure(err);
            }
        }
    }

    size_t read_some(size_t max, void* storage)
    {
        if (max == 0 || !storage) return 0;
        ssize_t res = recv(m_sockfd, storage, max, 0);
        if (res <= 0)
        {
            std::string err;
            if (res == 0) err = "Cannot read from a closed socket";
            else err = "Could not read from socket";
            HC_LOG_DEBUG(err);
            throw std::ios_base::failure(err);
        }
        return static_cast<size_t>(res);
    }

    void read(size_t buf_size, void* buf)
    {
        if (buf_size == 0 || !buf) return;
        size_t written = 0;
        while (written < buf_size)
        {
            written += read_some(buf_size - written,
                                 reinterpret_cast<char*>(buf) + written);
        }
    }

    void close()
    {
        if (!m_closed)
        {
            hamcast::util::closesocket(m_sockfd);
            m_closed = true;
        }
    }

    bool closed() const
    {
        return m_closed;
    }

    void flush () { }

};

} // namespace <anonymous>

namespace hamcast { namespace util {

socket_io::ptr socket_io::create(const char* host, boost::uint16_t port)
{
    return new native_socket_io(host, port);
}

socket_io::ptr socket_io::create(native_socket s)
{
    return new native_socket_io(s);
}

} } // namespace hamcast::ipc
