/******************************************************************************\
 *  _   ___     ____  __               _                                      *
 * | | | \ \___/ /  \/  | ___ __ _ ___| |_                                    *
 * | |_| |\     /| |\/| |/ __/ _` / __| __|                                   *
 * |  _  | \ - / | |  | | (_| (_| \__ \ |_                                    *
 * |_| |_|  \_/  |_|  |_|\___\__,_|___/\__|                                   *
 *                                                                            *
 * This file is part of the HAMcast project.                                  *
 *                                                                            *
 * HAMcast is free software: you can redistribute it and/or modify            *
 * it under the terms of the GNU Lesser General Public License as published   *
 * by the Free Software Foundation, either version 3 of the License, or       *
 * (at your option) any later version.                                        *
 *                                                                            *
 * HAMcast is distributed in the hope that it will be useful,                 *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of             *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                       *
 * See the GNU Lesser General Public License for more details.                *
 *                                                                            *
 * You should have received a copy of the GNU Lesser General Public License   *
 * along with HAMcast. If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                            *
 * Contact: HAMcast support <hamcast-support@informatik.haw-hamburg.de>       *
\******************************************************************************/

#include <cstdlib>      // abort()
#include <fstream>      // fstream
#include <iostream>     // cout, cerr, endl
#include <typeinfo>     // typeid(), type_info
#include <exception>    // set_terminate()

// mkdir
#include <signal.h>     // sigaction
#include <unistd.h>     // unlink
#include <sys/stat.h>   // mkdir, rmdir
#include <fcntl.h>      // creat()

#include <boost/regex.hpp>
#include <boost/thread.hpp>
#include <boost/cstdint.hpp>
#include <boost/date_time.hpp>
#include <boost/filesystem.hpp>
#include <boost/thread/thread_time.hpp>
#include <boost/interprocess/exceptions.hpp>
#include <boost/interprocess/sync/file_lock.hpp>
#include <boost/date_time/gregorian/gregorian.hpp>

#include "session.hpp"
#include "middleware.hpp"
#include "hamcast/hamcast_logging.h"

#include "hamcast/util/sink.hpp"
#include "hamcast/util/source.hpp"
#include "hamcast/util/socket_io.hpp"
#include "hamcast/util/buffered_sink.hpp"
#include "hamcast/util/buffered_source.hpp"
#include "hamcast/ipc/middleware_configuration.hpp"

#ifdef WIN_32
#else
#include <arpa/inet.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <cxxabi.h>         // abi::__cxa_demangle
#endif

using std::cout;
using std::cerr;
using std::endl;

using namespace hamcast;
using namespace hamcast::ipc;
using namespace hamcast::util;
using namespace hamcast::middleware;

typedef std::pair<source::ptr, sink::ptr> io_ptrs;

namespace {

typedef boost::uint16_t ui16;

bool verify_client(const source::ptr& in)
{
    HC_LOG_TRACE("");
    deserializer d(in);
    boost::uint32_t value;
    // the first value must be the magic number
    d >> value;
    if (value != magic_number)
    {
        HC_LOG_DEBUG("Wrong magic number");
        return false;
    }
    // the second value must be the major version of the client
    d >> value;
    if (value < min_compatile_major_version || value > major_version)
    {
        HC_LOG_DEBUG("Incompatible major version");
        return false;
    }
    // the third value must be the minor version of the client
    d >> value;
    if (value < min_compatile_minor_version || value > minor_version)
    {
        HC_LOG_DEBUG("Incompatible minor version");
        return false;
    }
    HC_LOG_DEBUG("New client accepted");
    return true;
}

bool client_compatible(const io_ptrs& io)
{
    HC_LOG_TRACE("");
    boost::uint8_t result = verify_client(io.first) ? 1 : 0;
    // lifetime scope of s
    {
        serializer s(io.second);
        s << result;
    }
    return result == 1;
}

void wait4clients_loop()
{
    HC_LOG_TRACE("");

    cout << "run loop ..." << endl;

    // start server at random port
    hamcast::util::native_socket ssock; // server socket
    hamcast::util::native_socket csock; // client socket
    HC_REQUIRE((ssock = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP)) != 0);
    struct sockaddr_in cl_addr; // client address
    memset(&cl_addr, 0, sizeof(sockaddr_in));
    int cl_len = sizeof(sockaddr_in);
    struct sockaddr_in sv_addr; // server address
    memset(&sv_addr, 0, sizeof(sockaddr_in));
    sv_addr.sin_family = AF_INET;
    sv_addr.sin_addr.s_addr = htonl(INADDR_ANY);
    for (;;)
    {
        // choose a random port > 1024
        ui16 port = static_cast<ui16>(((rand() % ((0xFFFF) - 1025)) + 1025));
        sv_addr.sin_port = htons(port);
        // try to bind to the local address
        if (bind(ssock, (struct sockaddr*) &sv_addr, sizeof(sockaddr_in)) == 0)
        {
            HC_LOG_DEBUG("Middleware runs on port " << port);
            // lifetime scope of mc
            {
                ipc::middleware_configuration mc(port);
                mc.write();
            }
            // max 10 pending connections
            HC_REQUIRE(listen(ssock, 10) == 0);
            for (;;)
            {
                HC_REQUIRE((csock = accept(ssock,
                                           (struct sockaddr*) &cl_addr,
                                           (socklen_t*) &cl_len)) > 0);
                socket_io::ptr sio(socket_io::create(csock));
                source::ptr in(new buffered_source<>(sio));
                sink::ptr out(new buffered_sink<>(sio));
                if (client_compatible(io_ptrs(in, out)))
                {
                    session::create(in, out);
                }
                else
                {
                    HC_LOG_DEBUG("Client rejected");
                }
            }
        }
    }
}

typedef std::map<std::string, string_map> config_map;

typedef void (*regex_handler)(boost::smatch&, config_map&, std::string&);
typedef std::pair<boost::regex, regex_handler> rx_handler_pair;
typedef std::vector<rx_handler_pair> rx_handler_t;

void push_back(rx_handler_t& v, const boost::regex& rx, regex_handler rxh)
{
    v.push_back(rx_handler_pair(rx, rxh));
}

bool matches(rx_handler_t& vec, const std::string& line,
             config_map& ini_map, std::string& section)
{
    boost::smatch sm;
    for (rx_handler_t::const_iterator i(vec.begin()); i != vec.end(); ++i)
    {
        if (boost::regex_match(line, sm, i->first))
        {
            (i->second)(sm, ini_map, section);
            return true;
        }
    }
    return false;
}

void new_section(boost::smatch& sm, config_map& cm, std::string& sec)
{
    std::string sec_name = sm[1];

    if (cm.find(sec_name) != cm.end())
    {
        std::string err;
        err  = "section ";
        err += sec_name;
        err += " already defined";
        throw std::runtime_error(err);
    }
    sec = sec_name;
    // insert sec_name to the map
    cm[sec_name];
}

void add_key_value_pair(boost::smatch& sm, config_map& cm, std::string& sec)
{
    if (sec.empty())
    {
        throw std::runtime_error("key value pair outside section");
    }
    std::string key = sm[1];
    std::string value = sm[2];
    string_map& kv_map = cm[sec];
    if (kv_map.find(key) != kv_map.end())
    {
        std::string err;
        err  = "key ";
        err += key;
        err += " already defined";
        throw std::runtime_error(err);
    }
    kv_map[key] = value;
}

void read_ini(config_map& conf)
{
    cout << "read config ..." << endl;
    // stores current section
    std::string section;
    // INI line handler
    rx_handler_t rx_handler;
    // INI section
    push_back(rx_handler,
              boost::regex("^\\[([a-zA-Z0-9\\._]+)\\]\\s*(?:;.*)?$"),
              new_section);
    // INI value
    push_back(rx_handler,
              boost::regex("^([a-zA-Z0-9\\._]+)\\s*="
                           "\\s*([a-zA-Z0-9\\._/]+)\\s*(?:;.*)?$"),
              add_key_value_pair);
    // INI string value
    push_back(rx_handler,
              boost::regex("^([a-zA-Z0-9\\._]+)\\s*="
                           "\\s*\"([^\"]+)\"\\s*(?:;.*)?$"),
              add_key_value_pair);
    int line_count = 0;
    std::string line;
    std::ifstream input_stream("middleware.ini");
    if (!(input_stream.is_open() && input_stream.good()))
    {
        throw std::runtime_error("Config file \"middleware.ini\" not found "
                                 "... quit");
    }
    while (std::getline(input_stream, line))
    {
        ++line_count;
        if (line.empty())
        {
            // ignore empty lines
        }
        else if (line.at(0) == ';')
        {
            // ignore comment lines
        }
        else if (!matches(rx_handler, line, conf, section))
        {
            cerr << "Could not parse line " << line_count
                 << " of middleware.ini (skipped)" << endl;
        }
    }
}

void set_log_level(int lvl)
{
    if (hc_logging_enabled())
    {
        hc_set_default_log_fun(lvl);
    }
    else
    {
        std::cout << "WARNING: log level ignored"
                     " (libHAMcast compiled without logging)"
                  << endl;
    }
}

boost::interprocess::file_lock* middleware_lock;//(lock_file.c_str());

int run_middleware()
{
    //boost::interprocess::file_lock middleware_lock(lock_file.c_str());
    //if (!middleware_lock.try_lock())
    //{
    //    cerr << "Only one instance of the middleware could be active." << endl
    //         << "If you are sure that there is no other instance is running "
    //         << "then delete the file \""
    //         << lock_file << "\"." << endl;
    //    return 2;
    //}
    // lifetime scope of ini
    {
        config_map ini;
        read_ini(ini);
        const string_map& global = ini["global"];
        for (string_map::const_iterator i = global.begin();
             i != global.end();
             ++i)
        {
            if (i->first == "log_level")
            {
                const std::string& lvl = i->second;
                if (lvl == "trace")
                {
                    set_log_level(HC_LOG_TRACE_LVL);
                }
                else if (lvl == "debug")
                {
                    set_log_level(HC_LOG_DEBUG_LVL);
                }
                else if (lvl == "info")
                {
                    set_log_level(HC_LOG_INFO_LVL);
                }
                else if (lvl == "warn")
                {
                    set_log_level(HC_LOG_WARN_LVL);
                }
                else if (lvl == "error")
                {
                    set_log_level(HC_LOG_ERROR_LVL);
                }
                else if (lvl == "fatal")
                {
                    set_log_level(HC_LOG_FATAL_LVL);
                }
                else if (lvl == "none")
                {
                    hc_set_log_fun(0);
                }
                else
                {
                    cerr << "illegal log level; expected one of "
                            "trace, debug, info, warn, error, fatal or none"
                         << endl;
                }
            }
            else
            {
                cerr << "unknown key found: \"" << i->first << "\" "
                        "(valid keys: log_level)"
                     << endl;
            }
        }
        for (config_map::iterator i = ini.begin(); i != ini.end(); ++i)
        {
            const std::string& section = i->first;
            if (section != "global")
            {
                const std::string& fname = (i->second)["file"];
                if (fname.empty())
                {
                    cerr << "no file name specified for module "
                         << section << endl;
                }
                else
                {
                    tech_module::load(fname.c_str(), i->second);
                }
            }
        }
        wait4clients_loop();
        return 0;
    }
}

volatile int exit_value = EXIT_SUCCESS;

void erase_trailing_slash(std::string& path)
{
    while (path.size() > 1 && *path.rbegin() == '/')
    {
        path.resize(path.size() - 1);
    }
}

void ensure_trailing_slash(std::string& path)
{
    if (path.empty())
    {
        throw std::runtime_error("path.emtpy()");
    }
    if (path.size() > 1)
    {
        if (*path.rbegin() != '/')
        {
            path += '/';
        }
    }
}

void remove_file(const std::string& cfile)
{
    if (boost::filesystem::exists(cfile.c_str()))
    {
        if (unlink(cfile.c_str()) != 0)
        {
            cerr << "could not remove "
                 << meeting_point << config_filename
                 << endl;
        }
    }
}

void cleanup()
{
    if (middleware_lock)
    {
        delete middleware_lock;
    }
    std::string path = meeting_point;
    std::string root = meeting_point_root;
    ensure_trailing_slash(path);
    // remove middleware config file
    std::string cfile = path;
    cfile += config_filename;
    remove_file(cfile);
    // remove lock file
    cfile = path;
    cfile += lock_filename;
    remove_file(cfile);
    erase_trailing_slash(path);
    while (path.size() > root.size())
    {
        if (rmdir(path.c_str()) == 0)
        {
            cout << "removed directory " << path << endl;
        }
        else
        {
            cout << "could not remove directory " << path << endl;
            return;
        }
        std::string::size_type new_size = path.find_last_of('/');
        path.resize(new_size);
    }
}

std::string demangle(const std::type_info& tinfo)
{
    size_t size;
    int status;
    char* undecorated = abi::__cxa_demangle(tinfo.name(), 0, &size, &status);
    if (status != 0)
    {
        return tinfo.name();
    }
    else
    {
        std::string result(undecorated, size);
        free(undecorated);
        return result;
    }
}

void print_exception(const std::exception& e)
{
    cerr << "terminate() called after throwing an instance of "
         << demangle(typeid(e))
         << "\nwhat(): " << e.what()
         << endl;
}

void terminate_handler()
{
    // throw exception again
    try { throw; }
    // and get the reason
    catch (std::exception& e)
    {
        print_exception(e);
    }
    catch (...)
    {
        cerr << "terminate() called after throwing an unknown exception"
             << endl;
    }
    cleanup();
    exit(3);
}

enum creation_result
{
    cr_ok,
    cr_exist,
    cr_failure
};

creation_result create_directory(const std::string& dir)
{
    if (mkdir(dir.c_str(), 0755) == 0)
    {
        return cr_ok;
    }
    if (errno == EEXIST)
    {
        return cr_exist;
    }
    else
    {
        return cr_failure;
    }
}

void create_file(const std::string& filename, bool throw_if_file_exists = false)
{
    int fhandle = creat(filename.c_str(), 0644);
    if (fhandle != -1)
    {
        close(fhandle);
    }
    else if (throw_if_file_exists || errno != EEXIST)
    {
        std::ostringstream errstr;
        errstr << "could not create " << filename;
        if (errno == EEXIST)
        {
            errstr << " (file already exists)";
        }
        else
        {
            errstr << " (error code: " << errno << ")";
        }
        throw std::runtime_error(errstr.str());
    }
}


void initialize()
{
    umask(0);
    std::string path = meeting_point;
    std::string root = meeting_point_root;
    if (path.empty())
    {
        throw std::logic_error("meeting_point is an empty string");
    }
    if (path.at(0) != '/')
    {
        throw std::logic_error("meeting_point is not an absolute path");
    }
    ensure_trailing_slash(path);
    std::string::size_type pos = path.find('/', root.size());
    for ( ; pos < path.size(); pos = path.find('/', pos + 1))
    {
        std::string sub = path.substr(0, pos);
        switch (create_directory(sub))
        {
         case cr_ok:
            cout << "created directory " << sub << endl;
            break;
         case cr_exist:
            cout << "direcory " << sub << " already exists" << endl;
            break;
         case cr_failure:
            cerr << "failure during creation of directory " << sub << endl;
            abort();
        }
    }
    // create config file
    std::string fname = path;
    fname += config_filename;
    create_file(fname);
    // create lock file
    fname = path;
    fname += lock_filename;
    create_file(fname, true);
    // get file lock
    middleware_lock = new boost::interprocess::file_lock(fname.c_str());
    if (!middleware_lock->try_lock())
    {
        throw std::runtime_error("Could not get interprocess file lock");
    }
}

void sigterm_handler(int)
{
    cout << endl;
    cleanup();
    exit(4);
}

void set_sigterm_handler()
{
    signal(SIGTERM, sigterm_handler);
    signal(SIGABRT, sigterm_handler);
    signal(SIGINT, sigterm_handler);
}

} // namespace <anonymous>

int main(int, char**)
{
    std::set_terminate(terminate_handler);
    set_sigterm_handler();
    initialize();
    // init stuff
    srand(time(NULL));
    int exit_reason = 0;
    try
    {
        run_middleware();
    }
    catch (boost::interprocess::interprocess_exception& e)
    {
        exit_reason = 1;
        cerr << "Only one instance of the middleware could be active." << endl
             << "If you are sure that there is no other instance is running "
             << "then delete the file \""
             << meeting_point << lock_filename << "\"." << endl;
        std::string err_msg = "unable to get interprocess lock: ";
        err_msg += e.what();
        print_exception(std::runtime_error(err_msg));
    }
    catch (std::exception& e)
    {
        exit_reason = 2;
        print_exception(e);
    }
    catch (...)
    {
        exit_reason = 5;
    }
    cleanup();
    return exit_reason;
}
